# Copyright (C) 2015, 2016 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Logging interface.
"""
from __future__ import print_function
import json
import sys

import labm8 as lab


def colourise(colour, *args):
    return "".join([colour] + list(args) + [Colours.RESET])


def printf(colour, *args, **kwargs):
    string = colourise(colour, *args)
    print(string, **kwargs)


def pprint(data, **kwargs):
    print(json.dumps(data, sort_keys=True, indent=2, separators=(",", ": ")),
          **kwargs)


def info(*args, **kwargs):
    print("[INFO  ]", *args, **kwargs)


def debug(*args, **kwargs):
    print("[DEBUG ]", *args, **kwargs)


def warn(*args, **kwargs):
    print("[WARN  ]", *args, **kwargs)


def error(*args, **kwargs):
    print("[ERROR ]", *args, **kwargs)


def fatal(*args, **kwargs):
    returncode = kwargs.pop("status", 1)
    error("fatal:", *args, **kwargs)
    lab.exit(returncode)


def prof(*args, **kwargs):
    """
    Print a profiling message.

    Profiling messages are intended for printing runtime performance
    data. They are prefixed by the "PROF" title.

    Arguments:

        *args, **kwargs: Message payload.
    """
    print("[PROF  ]", *args, **kwargs)


class Colours:
    """
    Shell escape colour codes.
    """
    RESET   = '\033[0m'
    GREEN   = '\033[92m'
    YELLOW  = '\033[93m'
    BLUE    = '\033[94m'
    RED     = '\033[91m'



# Copyright (C) 2017 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Lock file mechanism.
"""
from __future__ import print_function
# Use absolute paths for imports so as to prevent a conflict with the
# system "time" module.
from __future__ import absolute_import

import datetime
import os
import time

from labm8 import fs


class Error(Exception):
    pass


class UnableToAcquireLockError(Error):
    """ thrown if cannot acquire lock """
    def __init__(self, lock):
        self.lock = lock

    def __str__(self):
        path = self.lock.path
        pid = self.lock.pid
        date = self.lock.date
        return """\
Unable to acquire file lock owned by a different process.
Lock acquired by process {pid} on {date}.
Lock path: {path}""".format(**vars())


class UnableToReleaseLockError(Error):
    """ thrown if cannot release lock """
    def __init__(self, lock):
        self.lock = lock

    def __str__(self):
        path = self.lock.path
        pid = self.lock.pid
        date = self.lock.date
        return """\
Unable to release file lock owned by a different process.
Lock acquired by process {pid} on {date}.
Lock path: {path}""".format(**vars())


class LockFile:
    """
    A lock file.

    Attributes:
        path (str): Path of lock file.
    """
    def __init__(self, path):
        """
        Create a new directory lock.
        Arguments:
            path (str): Path to lock file.
        """
        self.path = fs.path(path)

    @property
    def pid(self):
        """
        The process ID of the lock. Value is None if lock is not claimed.
        """
        if fs.exists(self.path):
            with open(self.path) as infile:
                data = infile.read()
                components = data.split()
                pid = int(components[0])
                return pid
        else:
            return None

    @property
    def date(self):
        """
        The date that the lock was acquired. Value is None if lock is unclaimed.
        """
        if fs.exists(self.path):
            with open(self.path) as lockfile:
                data = lockfile.read()
                components = data.split()
                date = datetime.date.fromtimestamp(float(components[1]))
                return date
        else:
            return None

    @property
    def islocked(self):
        """
        Whether the directory is locked.
        """
        return fs.exists(self.path)

    @property
    def owned_by_self(self):
        """
        Whether the directory is locked by the current process.
        """
        return self.pid == os.getpid()

    def acquire(self, force=False):
        """
        Acquire the lock.

        A lock can be claimed if any of these conditions are true:
            1. The lock is unheld by anyone.
            2. The lock is held but the 'force' argument is set.
            3. The lock is held by the current process.

        Arguments:
            force (boolean, optional): If true, ignore any existing
              lock. If false, fail if lock already claimed.

        Returns:
            LockFile: self.

        Raises:
            UnableToAcquireLockError: If the lock is already claimed
              (not raised if force option is used).
        """
        if not self.islocked or force or self.pid == os.getpid():
            with open(self.path, "w") as lockfile:
                print(os.getpid(), time.time(), file=lockfile)
            return self
        else:
            raise UnableToAcquireLockError(self)

    def release(self, force=False):
        """
        Release lock.

        To release a lock, we must already own the lock.

        Arguments:
            force (bool, optional): If true, ignore any existing lock owner.

        Raises:
            UnableToReleaseLockError: If the lock is claimed by another
              process (not raised if force option is used).
        """
        # There's no lock, so do nothing.
        if not self.islocked:
            return

        if self.owned_by_self or force:
            os.remove(self.path)
        else:
            raise UnableToReleaseLockError(self)

    def __repr__(self):
        return self.path

    def __enter__(self):
        return self.acquire()

    def __exit__(self, type, value, tb):
        self.release()



# Copyright (C) 2015, 2016 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
High level filesystem interface.
"""
import os
import os.path
import re
import shutil

from glob import iglob
from humanize import naturalsize
from send2trash import send2trash

import labm8 as lab
from labm8 import types


class Error(Exception):
    pass


class File404(Error):
    pass


def path(*components):
    """
    Get a file path.

    Concatenate all components into a path.
    """
    _path = os.path.join(*components)
    _path = os.path.expanduser(_path)

    return _path


def must_exist(*components):
    """
    Ensure path exists.

    Arguments:
        *components (str[]): Path components.

    Returns:
        str: File path.

    Raises:
        File404: If path does not exist.
    """
    _path = path(*components)
    if not exists(_path):
        raise File404(_path)
    return _path


def abspath(*components):
    """
    Get an absolute file path.

    Concatenate all components into an absolute path.
    """
    return os.path.abspath(path(*components))


def basename(*components):
    """
    Return the basename of a given file path.
    """
    return os.path.basename(path(*components))


def dirname(*components):
    """
    Return the directory name of a given file path.
    """
    return os.path.dirname(path(*components))


def is_subdir(child, parent):
    """
    Determine if "child" is a subdirectory of "parent".

    If child == parent, returns True.
    """
    child_path = os.path.realpath(child)
    parent_path = os.path.realpath(parent)

    if len(child_path) < len(parent_path):
        return False

    for i in range(len(parent_path)):
        if parent_path[i] != child_path[i]:
            return False

    return True


# Directory history.
_cdhist = []


def cd(path):
    """
    Change working directory.

    Returns absolute path to new working directory.
    """
    _cdhist.append(pwd())  # Push to history.
    path = abspath(path)
    os.chdir(path)
    return path


def cdpop():
    """
    Return the last directory.

    Returns absolute path to new working directory.
    """
    if len(_cdhist) >= 1:
        old = _cdhist.pop()  # Pop from history.
        os.chdir(old)
        return old
    else:
        return pwd()


def pwd():
    """
    Return the path to the current working directory.
    """
    return os.getcwd()


def exists(*components):
    """
    Return whether a file exists.
    """
    return os.path.exists(path(*components))


def isfile(*components):
    """
    Return whether a path exists, and is a file.
    """
    return os.path.isfile(path(*components))


def isexe(*components):
    """
    Return whether a path is an executable file.

    Arguments:

        path (str): Path of the file to check.

    Examples:

        >>> fs.isexe("/bin/ls")
        True

        >>> fs.isexe("/home")
        False

        >>> fs.isexe("/not/a/real/path")
        False

    Returns:

        bool: True if file is executable, else false.
    """
    _path = path(*components)
    return isfile(_path) and os.access(_path, os.X_OK)


def isdir(*components):
    """
    Return whether a path exists, and is a directory.
    """
    return os.path.isdir(path(*components))


def ls(root=".", abspaths=False, recursive=False):
    """
    Return a list of files in directory.

    Directory listings are sorted alphabetically. If the named
    directory is a file, return it's path.

    Examples:

        >>> fs.ls("foo")
        ["a", "b", "c"]

        >>> fs.ls("foo/a")
        ["foo/a"]

        >>> fs.ls("foo", abspaths=True)
        ["/home/test/foo/a", "/home/test/foo/b", "/home/test/foo/c"]

        >>> fs.ls("foo", recursive=True)
        ["a", "b", "b/d", "b/d/e", "c"]

    Arguments:

        root (str): Path to directory. Can be relative or absolute.
        abspaths (bool, optional): Return absolute paths if true.
        recursive (bool, optional): Recursively list subdirectories if
          true.

    Returns:

        list of str: A list of paths.

    Raises:

        OSError: If root directory does not exist.
    """
    def _expand_subdirs(file):
        if isdir(path(root, file)):
            return [file] + [path(file, x) for x in
                             ls(path(root, file), recursive=True)]
        else:
            return [file]

    if isfile(root):
        # If argument is a file, return path.
        return [abspath(root)] if abspaths else [basename(root)]
    elif abspaths:
        # Get relative names.
        relpaths = ls(root, recursive=recursive, abspaths=False)
        # Prepend the absolute path to each relative name.
        base = abspath(root)
        return [path(base, relpath) for relpath in relpaths]
    elif recursive:
        # Recursively expand subdirectories.
        paths = ls(root, abspaths=abspaths, recursive=False)
        return types.flatten([_expand_subdirs(file) for file in paths])
    else:
        # List directory contents.
        return list(sorted(os.listdir(root)))


def lsdirs(root=".", **kwargs):
    """
    Return only subdirectories from a directory listing.

    Arguments:

        root (str): Path to directory. Can be relative or absolute.
        **kwargs: Any additional arguments to be passed to ls().

    Returns:

        list of str: A list of directory paths.

    Raises:

        OSError: If root directory does not exist.
    """
    paths = ls(root=root, **kwargs)
    if isfile(root):
        return []
    return [_path for _path in paths if isdir(path(root, _path))]


def lsfiles(root=".", **kwargs):
    """
    Return only files from a directory listing.

    Arguments:

        root (str): Path to directory. Can be relative or absolute.
        **kwargs: Any additional arguments to be passed to ls().

    Returns:

        list of str: A list of file paths.

    Raises:

        OSError: If root directory does not exist.
    """
    paths = ls(root=root, **kwargs)
    if isfile(root):
        return paths
    return [_path for _path in paths if isfile(path(root, _path))]


def rm(*components, **kwargs):
    """
    Remove a file or directory.

    If path is a directory, this recursively removes the directory and
    any contents. Non-existent paths are silently ignored.

    Supports Unix style globbing by default (disable using
    glob=False). For details on globbing pattern expansion, see:

        https://docs.python.org/2/library/glob.html

    Arguments:
        *components (string[]): path to the file or directory to remove. May be
          absolute or relative. May contain unix glob
        **kwargs: if "glob" is True, perform Unix style pattern expansion of
          paths (default: True).
    """
    _path = path(*components)
    glob = kwargs.get("glob", True)
    paths = iglob(_path) if glob else [_path]

    for file in paths:
        if isfile(file):
            os.remove(file)
        elif exists(file):
            shutil.rmtree(file, ignore_errors=False)


def rmtrash(*components):
    """
    Move a file or directory to trash.

    If file does not exist, nothing happens.

    Examples:

        >>> fs.rmtrash("foo", "bar")

        >>> fs.rmtrash("/home/labm8/file.txt")

    Arguments:
        *components (string[]): path to the file or directory.
    """
    _path = path(*components)
    if exists(_path):
        send2trash(_path)


def cp(src, dst):
    """
    Copy a file or directory.

    If source is a directory, this recursively copies the directory
    and its contents. If the destination is a directory, then this
    creates a copy of the source in the destination directory with the
    same basename.

    If the destination already exists, this will attempt to overwrite
    it.

    Arguments:

        src (string): path to the source file or directory.
        dst (string): path to the destination file or directory.

    Raises:

        IOError: if source does not exist.
    """
    if isdir(src):
        # Overwrite an existing directory.
        if isdir(dst):
            rm(dst)
        shutil.copytree(src, dst)
    elif isfile(src):
        shutil.copy(src, dst)
    else:
        raise IOError("Source '{0}' not found".format(src))


def mv(src, dst):
    """
    Move a file or directory.

    If the destination already exists, this will attempt to overwrite
    it.

    Arguments:

        src (string): path to the source file or directory.
        dst (string): path to the destination file or directory.

    Raises:

        File404: if source does not exist.
        IOError: in case of error.
    """
    if not exists(src):
        raise File404(src)

    try:
        shutil.move(src, dst)
    except Exception as e:
        raise IOError(str(e))


def mkdir(*components, **kwargs):
    """
    Make directory "path", including any required parents. If
    directory already exists, do nothing.
    """
    _path = path(*components)
    if not isdir(_path):
        os.makedirs(_path, **kwargs)
    return _path


def mkopen(p, *args, **kwargs):
    """
    A wrapper for the open() builtin which makes parent directories if needed.
    """
    dir = os.path.dirname(p)
    mkdir(dir)
    return open(p, *args, **kwargs)


def read(*components, **kwargs):
    """
    Read file and return a list of lines. If comment_char is set, ignore the
    contents of lines following the comment_char.

    Raises:

        IOError: if reading path fails
    """
    rstrip = kwargs.get("rstrip", True)
    comment_char = kwargs.get("comment_char", None)

    ignore_comments = comment_char is not None

    file = open(path(*components))
    lines = file.readlines()
    file.close()

    # Multiple definitions to handle all cases.
    if ignore_comments:
        comment_line_re = re.compile("^\s*{char}".format(char=comment_char))
        not_comment_re = re.compile("[^{char}]+".format(char=comment_char))

        if rstrip:
            # Ignore comments, and right strip results.
            return [re.match(not_comment_re, line).group(0).rstrip()
                    for line in lines
                    if not re.match(comment_line_re, line)]
        else:
            # Ignore comments, and don't strip results.
            return [re.match(not_comment_re, line).group(0)
                    for line in lines
                    if not re.match(comment_line_re, line)]
    elif rstrip:
        # No comments, and right strip results.
        return [line.rstrip() for line in lines]
    else:
        # Just a good old-fashioned read!
        return lines


def du(*components, **kwargs):
    """
    Get the size of a file in bytes or as a human-readable string.

    Arguments:

        *components (str[]): Path to file.
        **kwargs: If "human_readable" is True, return a formatted string,
          e.g. "976.6 KiB" (default True)

    Returns:
        int or str: If "human_readble" kwarg is True, return str, else int.
    """
    human_readable = kwargs.get("human_readable", True)

    _path = path(*components)
    if not exists(_path):
        raise Error("file '{}' not found".format(_path))
    size = os.stat(_path).st_size
    if human_readable:
        return naturalsize(size)
    else:
        return size


def write_file(path, contents):
    """
    Write string to file.

    Arguments:
        path (str): Destination.
        contents (str): Contents.
    """
    with mkopen(path, 'w') as outfile:
        outfile.write(contents)


def read_file(path):
    """
    Read file to string.

    Arguments:
        path (str): Source.
    """
    with open(must_exist(path)) as infile:
        r = infile.read()
    return r


def files_from_list(*paths):
    """
    Return a list of all file paths from a list of files or directories.

    For each path in the input: if it is a file, return it; if it is a
    directory, return a list of files in the directory.

    Arguments:
        paths (list of str): List of file and directory paths.

    Returns:
        list of str: Absolute file paths.

    Raises:
        File404: If any of the paths do not exist.
    """
    ret = []
    for path in paths:
        if isfile(path):
            ret.append(abspath(path))
        elif isdir(path):
            ret += [f for f in ls(path, abspaths=True, recursive=True)
                    if isfile(f)]
        else:
            raise File404(path)
    return ret



# Copyright (C) 2015, 2016 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Transient and persistent caching mechanisms.
"""
import atexit
import json
import re
import six

import labm8 as lab
from labm8 import crypto
from labm8 import fs
from labm8 import io


class Cache(object):
    """
    Cache for storing (key,value) relational data.

    A cache is a dictionary with a limited subset of a the
    functionality.
    """

    def get(self, key, default=None):
        """
        Retrieve an item from cache.

        Arguments:
            key: Item key.
            default (optional): Default value if item not found.
        """
        raise NotImplementedError

    def clear(self):
        """
        Remove all items from cache.
        """
        raise NotImplementedError

    def items(self):
        """
        Returns a generator for iterating over (key, value) pairs.
        """
        raise NotImplementedError

    def __getitem__(self, key):
        """
        Retrieve an item from cache.

        Arguments:
           key: Item key.

        Raises:
           KeyError: If key is not in cache.
        """
        raise NotImplementedError

    def __setitem__(self, key, value):
        """
        Set (key, value) pair.
        """
        raise NotImplementedError

    def __contains__(self, key):
        """
        Returns whether key is in cache.
        """
        raise NotImplementedError

    def __delitem__(self, key):
        """
        Remove (key, value) pair.
        """
        raise NotImplementedError


class TransientCache(Cache):
    """
    An in-memory only cache.
    """

    def __init__(self, basecache=None):
        """
        Create a new transient cache.

        Optionally supports populating the cache with values of an
        existing cache.

        Arguments:
           basecache (TransientCache, optional): Cache to populate this new
             cache with.
        """
        self._data = {}

        if basecache is not None:
            for key,val in basecache.items():
                self._data[key] = val

    def get(self, key, default=None):
        if key in self._data:
            return self._data[key]
        else:
            return default

    def clear(self):
        self._data.clear()

    def items(self):
        return six.iteritems(self._data)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value
        return value

    def __contains__(self, key):
        return key in self._data

    def __delitem__(self, key):
        del self._data[key]


class JsonCache(TransientCache):
    """
    A persistent, JSON-backed cache.

    Requires that (key, value) pairs are JSON serialisable.
    """

    def __init__(self, path, basecache=None):
        """
        Create a new JSON cache.

        Optionally supports populating the cache with values of an
        existing cache.

        Arguments:
           basecache (TransientCache, optional): Cache to populate this new
             cache with.
        """

        super(JsonCache, self).__init__()
        self.path = fs.abspath(path)

        if fs.exists(self.path):
            io.debug(("Loading cache '{0}'".format(self.path)))
            with open(self.path) as file:
                self._data = json.load(file)

        if basecache is not None:
            for key,val in basecache.items():
                self._data[key] = val

        # Register exit handler
        atexit.register(self.write)

    def write(self):
        """
        Write contents of cache to disk.
        """
        io.debug("Storing cache '{0}'".format(self.path))
        with open(self.path, "w") as file:
            json.dump(self._data, file, sort_keys=True, indent=2,
                      separators=(',', ': '))


def hash_key(key):
    """
    Convert a key to a filename by hashing its value.
    """
    return crypto.sha1_str(json.dumps(key, sort_keys=True))


def escape_path(key):
    """
    Convert a key to a filename by escaping invalid characters.
    """
    return re.sub(r'[ \\/]+', '_', key)


class FSCache(Cache):
    """
    Persistent filesystem cache.

    Each key uniquely identifies a file.
    Each value is a file path.

    Adding a file to the cache moves it into the cahce directory.

    Members:
        path (str): Root cache.
        escape_key (fn): Function to convert keys to file names.
    """
    def __init__(self, root, escape_key=hash_key):
        """
        Create filesystem cache.

        Arguments:
            root (str): String.
            escape_key (fn, optional): Function to convert keys to file names.
        """
        self.path = root
        self.escape_key = escape_key

        fs.mkdir(self.path)

    def clear(self):
        """
        Empty the filesystem cache.

        This deletes the entire cache directory.
        """
        fs.rm(self.path)

    def keypath(self, key):
        """
        Get the filesystem path for a key.

        Arguments:
            key: Key.

        Returns:
            str: Absolute path.
        """
        return fs.path(self.path, self.escape_key(key))

    def __getitem__(self, key):
        """
        Get path to file in cache.

        Arguments:
            key: Key.

        Returns:
            str: Path to cache value.

        Raises:
            KeyErorr: If key not in cache.
        """
        path = self.keypath(key)
        if fs.exists(path):
            return path
        else:
            raise KeyError(key)

    def __setitem__(self, key, value):
        """
        Emplace file in cache.

        Arguments:
            key: Key.
            value (str): Path of file to insert in cache.

        Raises:
            ValueError: If no "value" does nto exist.
        """
        if not fs.exists(value):
            raise ValueError(value)

        path = self.keypath(key)
        fs.mv(value, path)

    def __contains__(self, key):
        """
        Check cache contents.

        Arguments:
            key: Key.

        Returns:
            bool: True if key in cache, else false.
        """
        path = self.keypath(key)
        return fs.exists(path)

    def __delitem__(self, key):
        """
        Delete cached file.

        Arguments:
            key: Key.

        Raises:
            KeyError: If file not in cache.
        """
        path = self.keypath(key)
        if fs.exists(path):
            fs.rm(path)
        else:
            raise KeyError(key)

    def get(self, key, default=None):
        """
        Fetch from cache.

        Arguments:
            key: Key.
            default (optional): Value returned if key not found.

        Returns:
            str: Path to cached file.
        """
        if key in self:
            return self[key]
        else:
            return default



# Copyright (C) 2015, 2016 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Utilies for grokking the underlying system.

Variables:
  * `HOSTNAME` (str) System hostname.
  * `USERNAME` (str) Username.
  * `UID` (int) User ID.
  * `PID` (int) Process ID.
"""

from __future__ import print_function

import getpass
import os
import socket
import sys
import subprocess
import threading

from sys import platform

import labm8 as lab
from labm8 import fs
from labm8 import io

HOSTNAME = socket.gethostname()
USERNAME = getpass.getuser()
UID = os.getuid()
PID = os.getpid()

argv = sys.argv
STDOUT = sys.stdout
STDERR = sys.stderr
PIPE = subprocess.PIPE


class Error(Exception):
    pass


class SubprocessError(Error):
    """
    Error thrown if a subprocess fails.
    """
    pass


class CommandNotFoundError(Exception):
    """
    Error thrown a system command is not found.
    """
    pass


class ScpError(Error):
    """
    Error thrown if scp file transfer fails.
    """
    def __init__(self, stdout, stderr):
        """
        Construct an ScpError.

        Arguments:

            stdout (str): Captured stdout of scp subprocess.
            stderr (str): Captured stderr of scp subprocess.
        """
        self.out = stdout
        self.err = stderr

    def __repr__(self):
        return self.out + "\n" + self.err

    def __str__(self):
        return self.__repr__()


class Subprocess(object):
    """
    Subprocess abstraction.

    Wrapper around subprocess.Popen() which provides the ability to
    force a timeout after a number of seconds have elapsed.
    """
    def __init__(self, cmd, shell=False,
                 stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE,
                 decode_out=True):
        """
        Create a new subprocess.
        """
        self.cmd = cmd
        self.process = None
        self.stdout = None
        self.stderr = None
        self.shell = shell
        self.decode_out = decode_out

        self.stdout_dest = stdout
        self.stderr_dest = stderr

    def run(self, timeout=-1):
        """
        Run the subprocess.

        Arguments:
            timeout (optional) If a positive real value, then timout after
                the given number of seconds.

        Raises:
            SubprocessError If subprocess has not completed after "timeout"
                seconds.
        """
        def target():
            self.process = subprocess.Popen(self.cmd,
                                            stdout=self.stdout_dest,
                                            stderr=self.stderr_dest,
                                            shell=self.shell)
            stdout, stderr = self.process.communicate()

            # Decode output if the user wants, and if there is any.
            if self.decode_out:
                if stdout:
                    self.stdout = stdout.decode("utf-8")
                if stderr:
                    self.stderr = stderr.decode("utf-8")

        thread = threading.Thread(target=target)
        thread.start()

        if timeout > 0:
            thread.join(timeout)
            if thread.is_alive():
                self.process.terminate()
                thread.join()
                raise SubprocessError(("Reached timeout after {t} seconds"
                                       .format(t=timeout)))
        else:
            thread.join()

        return self.process.returncode, self.stdout, self.stderr


def is_linux():
    return platform == "linux" or platform == "linux2"


def is_mac():
    return platform == "darwin"


def is_windows():
    return platform == "win32"


def run(command, num_retries=1, timeout=-1, **kwargs):
    """
    Run a command with optional timeout and retries.

    Provides a convenience method for executing a subprocess with
    additional error handling.

    Arguments:
        command (list of str): The command to execute.
        num_retries (int, optional): If the subprocess fails, the number of
          attempts to execute it before failing.
        timeout (float, optional): If positive, the number of seconds to wait
          for subprocess completion before failing.
        **kwargs: Additional args to pass to Subprocess.__init__()

    Returns:
        Tuple of (int, str, str): Where the variables represent
        (exit status, stdout, stderr).

    Raises:
        SubprocessError: If the command fails after the given number of
          retries.
    """
    last_error = None
    for _ in range(num_retries):
        try:
            process = Subprocess(command, **kwargs)
            return process.run(timeout)
        except Exception as err:
            last_error = err

    raise last_error


def sed(match, replacement, path, modifiers=""):
    """
    Perform sed text substitution.
    """
    cmd = "sed -r -i 's/%s/%s/%s' %s" % (match, replacement, modifiers, path)

    process = Subprocess(cmd, shell=True)
    ret, out, err = process.run(timeout=60)
    if ret:
        raise SubprocessError("Sed command failed!")


def echo(*args, **kwargs):
    """
    Write a message to a file.

    Arguments:
        args A list of arguments which make up the message. The last argument
            is the path to the file to write to.
    """
    msg = args[:-1]
    path = fs.path(args[-1])
    append = kwargs.pop("append", False)

    if append:
        with open(path, "a") as file:
            print(*msg, file=file, **kwargs)
    else:
        with open(fs.path(path), "w") as file:
            print(*msg, file=file, **kwargs)


def which(program, path=None):
    """
    Returns the full path of shell commands.

    Replicates the functionality of system which (1) command. Looks
    for the named program in the directories indicated in the $PATH
    environment variable, and returns the full path if found.

    Examples:

        >>> system.which("ls")
        "/bin/ls"

        >>> system.which("/bin/ls")
        "/bin/ls"

        >>> system.which("not-a-real-command")
        None

        >>> system.which("ls", path=("/usr/bin", "/bin"))
        "/bin/ls"

    Arguments:

        program (str): The name of the program to look for. Can
          be an absolute path.
        path (sequence of str, optional): A list of directories to
          look for the pgoram in. Default value is system $PATH.

    Returns:

       str: Full path to program if found, else None.
    """
    # If path is not given, read the $PATH environment variable.
    path = path or os.environ["PATH"].split(os.pathsep)
    abspath = True if os.path.split(program)[0] else False
    if abspath:
        if fs.isexe(program):
            return program
    else:
        for directory in path:
            # De-quote directories.
            directory = directory.strip('"')
            exe_file = os.path.join(directory, program)
            if fs.isexe(exe_file):
                return exe_file

    return None


def scp(host, src, dst, user=None, path=None):
    """
    Copy a file or directory from a remote location.

    A thin wrapper around the scp (1) system command.

    If the destination already exists, this will attempt to overwrite
    it.

    Arguments:

        host (str): name of the host
        src (str): path to the source file or directory.
        dst (str): path to the destination file or directory.
        user (str, optional): Alternative username for remote access.
          If not provided, the default scp behaviour is used.
        path (str, optional): Directory containing scp command. If not
          provided, attempt to locate scp using which().

    Raises:

        CommandNotFoundError: If scp binary not found.
        IOError: if transfer fails.
    """
    # Create the first argument.
    if user is None:
        arg = "{host}:{path}".format(host=host, path=src)
    else:
        arg = "{user}@{host}:{path}".format(user=user, host=host, path=src)

    # Get path to scp binary.
    scp_bin = which("scp", path=(path,))
    if scp_bin is None:
        raise CommandNotFoundError("Could not find scp in '{0}'".format(path))

    # Run system "scp" command.
    ret,out,err = run([scp_bin,
                       "-o", "StrictHostKeyChecking=no",
                       "-o", "UserKnownHostsFile=/dev/null",
                       arg, dst])

    # Check return code for error.
    if ret:
        raise ScpError(out, err)


def isprocess(pid, error=False):
    """
    Check that a process is running.

    Arguments:

        pid (int): Process ID to check.

    Returns:

        True if the process is running, else false.
    """
    try:
        # Don't worry folks, no processes are harmed in the making of
        # this system call:
        os.kill(pid, 0)
        return True
    except OSError:
        return False



# Copyright (C) 2015, 2016 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Utils for manipulating quantitative experimental data.
"""
import labm8 as lab
import labm8.io

import sys

from pkg_resources import require


__version__ = require("labm8")[0].version


def exit(status=0):
    """
    Terminate the program with the given status code.
    """
    if status == 0:
        lab.io.printf(lab.io.Colours.GREEN, "Done.")
    else:
        lab.io.printf(lab.io.Colours.RED, "Error {0}".format(status))
    sys.exit(status)


def is_python3():
    """
    Returns whether the Python version is >= 3.0.

    This is for compatability purposes, where you need to implement different
    code for Python 2 and 3.

    Example:
        To import the StringIO class:

          if is_python3():
            from io import StringIO
          else:
            from StringIO import StringIO

    Returns:
        bool: True if Python >= 3, else False.
    """
    return sys.version_info >= (3, 0)



# Copyright (C) 2015, 2016 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Text utilities.
"""
from __future__ import division

import re

import labm8 as lab


class Error(Exception):
    """
    Module-level error.
    """
    pass


class TruncateError(Error):
    """
    Thrown in case of truncation error.
    """
    pass


def get_substring_idxs(substr, string):
    """
    Return a list of indexes of substr. If substr not found, list is
    empty.

    Arguments:
        substr (str): Substring to match.
        string (str): String to match in.

    Returns:
        list of int: Start indices of substr.
    """
    return [match.start() for match in re.finditer(substr, string)]


def truncate(string, maxchar):
    """
    Truncate a string to a maximum number of characters.

    If the string is longer than maxchar, then remove excess
    characters and append an ellipses.

    Arguments:

        string (str): String to truncate.
        maxchar (int): Maximum length of string in characters. Must be >= 4.

    Returns:

        str: Of length <= maxchar.

    Raises:

        TruncateError: In case of an error.
    """
    if maxchar < 4:
        raise TruncateError("Maxchar must be > 3")

    if len(string) <= maxchar:
        return string
    else:
        return string[:maxchar - 3] + "..."


def levenshtein(s1, s2):
    """
    Return the Levenshtein distance between two strings.

    Implementation of Levenshtein distance, one of a family of edit
    distance metrics.

    Based on: https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python

    Examples:

        >>> text.levensthein("foo", "foo")
        0

        >>> text.levensthein("foo", "fooo")
        1

        >>> text.levensthein("foo", "")
        3

        >>> text.levensthein("1234", "1 34")
        1

    Arguments:

        s1 (str): Argument A.
        s2 (str): Argument B.

    Returns:

        int: Levenshtein distance between the two strings.
    """
    # Left string must be >= right string.
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    # Distance is length of s1 if s2 is empty.
    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def diff(s1, s2):
    """
    Return a normalised Levenshtein distance between two strings.

    Distance is normalised by dividing the Levenshtein distance of the
    two strings by the max(len(s1), len(s2)).

    Examples:

        >>> text.diff("foo", "foo")
        0

        >>> text.diff("foo", "fooo")
        1

        >>> text.diff("foo", "")
        1

        >>> text.diff("1234", "1 34")
        1

    Arguments:

        s1 (str): Argument A.
        s2 (str): Argument B.

    Returns:

        float: Normalised distance between the two strings.
    """
    return levenshtein(s1, s2) / max(len(s1), len(s2))



# Copyright (C) 2015, 2016 Chris Cummins.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
"""
Profiling API for timing critical paths in code.
"""
from __future__ import absolute_import

import random

from time import time

import labm8 as lab
from labm8 import fs
from labm8 import io


_GLOBAL_TIMER = "__global__"
_timers = {}


class Error(Exception):
    """
    Module-level error class.
    """
    pass


class TimerNameError(Error):
    """
    Thrown in case of timer name conflicts or lookup misses.
    """
    pass


def isrunning(name=None):
    """
    Check if a timer is running.

    Arguments:

        name (str, optional): The name of the timer to check.

    Returns:

        bool: True if timer is running, else False.
    """
    name = name or _GLOBAL_TIMER
    return name in _timers


def _add_timer(name):
    _timers[name] = time()


def _get_elapsed_time(name):
    t = _timers[name]
    return (time() - t) * 1000


def _stop_timer(name):
    elapsed = _get_elapsed_time(name)
    del _timers[name]
    return elapsed


def _new_timer_name():
    name = "{:08x}".format(random.randrange(16 ** 8))
    return name if not isrunning(name) else _new_timer_name()


def start(name=None, unique=False):
    """
    Start a new profiling timer.

    Arguments:

        name (str, optional): The name of the timer to create. If no
          name is given, the resulting timer is anonymous. There can
          only be one anonymous timer.
        unique (bool, optional): If true, then ensure that timer name
          is unique. This is to prevent accidentally resetting an
          existing timer.

    Raises:

        TimerNameError: If `unique' is true and a timer with
          the same name already exists.
    """
    name = name or _GLOBAL_TIMER

    if unique and isrunning(name):
        raise TimerNameError("A timer named '{}' already exists".format(name))

    _add_timer(name)


def new():
    """
    Create and start a uniquely named timer.

    Generates a timer name which is guaranteed to be unique, then
    starts it.

    Returns:

        str: Name of new timer.
    """
    name = _new_timer_name()
    start(name=name)
    return name


def stop(name=None, **kwargs):
    """
    Stop a timer.

    Arguments:

       name (str, optional): The name of the timer to stop. If no name
         is given, stop the global anonymous timer.

    Raises:

        TimerNameError: If the named timer does not exist.
    """
    name = name or _GLOBAL_TIMER

    if name not in _timers:
        if name == _GLOBAL_TIMER:
            raise Error("Global timer has not been started")
        else:
            raise TimerNameError("No timer named '{}'".format(name))

    elapsed = int(round(_stop_timer(name)))
    if name == _GLOBAL_TIMER:
        name = "Timer"

    io.prof("{}: {} ms".format(name, elapsed), **kwargs)


def end(*args, **kwargs):
    """
    Stop a timer, see stop().
    """
    stop(*args, **kwars)


def reset(name=None):
    """
    Reset a timer.

    Arguments:

        name (str, optional): The name of the timer to reset. If no
          name is given, reset the global anonymous timer.

    Raises:

        TimerNameError: If the named timer does not exist.
    """
    if name not in _timers:
        raise TimerNameError("No timer named '{}'".format(name))

    _add_timer(name)


def elapsed(name=None):
    name = name or _GLOBAL_TIMER

    if name not in _timers:
        if name == _GLOBAL_TIMER:
            raise Error("Global timer has not been started")
        else:
            raise TimerNameError("No timer named '{}'".format(name))

    return _get_elapsed_time(name)



# Copyright (C) 2015, 2016 Chris Cummins.
#
# This file is part of labm8.
#
# Labm8 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# Labm8 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with labm8.  If not, see <http://www.gnu.org/licenses/>.
"""
Graphing helper.
"""
import labm8 as lab
from labm8 import io


class Error(Exception):
    """
    Visualisation module base error class.
    """
    pass


def finalise(output=None, figsize=None, tight=True, **kwargs):
    """
    Finalise a plot.

    Display or show the plot, then close it.

    Arguments:

        output (str, optional): Path to save figure to. If not given,
          show plot.
        figsize ((float, float), optional): Figure size in inches.
        **kwargs: Any additional arguments to pass to
          plt.savefig(). Only required if output is not None.
    """
    import matplotlib.pyplot as plt

    # Set figure size.
    if figsize is not None:
        plt.gcf().set_size_inches(*figsize)

    # Set plot layout.
    if tight:
        plt.tight_layout()

    if output is None:
        plt.show()
    else:
        plt.savefig(output, **kwargs)
        io.info("Wrote", output)
    plt.close()
